package i5d.gui;

import java.awt.*;
import java.awt.event.*;

import i5d.Image5D;
import ij.*;
import ij.gui.*;
import ij.measure.*;

/*
 * Created on 28.03.2005
*/

/** ImageWindow for Image5Ds. Has two scrollbars for slice and time and a panel with controls
 * to change the current channel and its color.
 * @author Joachim Walter
 */
public class Image5DWindow extends StackWindow implements KeyListener {

    private static final long serialVersionUID = -3196514227677416036L;
    
    protected ChannelControl channelControl;
    protected ScrollbarWithLabel[] scrollbarsWL;
	protected Scrollbar[] Scrollbars;
	protected Image5D i5d;
	
	// Array for storing change of position in each dimension. 
	// 0: no change, 1 - dimensionSize : changed position
	protected int[] positions;
	
	protected int nDimensions = 5;
    protected int[] dimensions;
	
	protected boolean isInitialized = false;
	
	/**
	 * @param imp
	 */
	public Image5DWindow(Image5D imp) {
		this(imp, new Image5DCanvas(imp));
	}

	/**
	 * @param imp
	 * @param ic
	 */
	public Image5DWindow(Image5D imp, Image5DCanvas ic) {
		super(imp, ic);
		
		if (ic == null) {
			throw new IllegalArgumentException("Image5DCanvas must not be null.");
		}
        
		i5d = (Image5D)imp;

		if (imp.getNDimensions() != nDimensions) {
			throw new IllegalArgumentException("Wrong number of dimensions.");
		}

        scrollbarsWL = new ScrollbarWithLabel[nDimensions];
        Scrollbars = new Scrollbar[nDimensions];
		positions = new int[nDimensions];
        dimensions = i5d.getDimensions();

		// Remove all components and then add them with the Image5DLayout layoutmanager.
		remove(sliceSelector);
		remove(ic);

		setLayout(new Image5DLayout(ic));
		
		// Add ImageCanvas
		add(ic, Image5DLayout.CANVAS);		

		// Add channel selector
		channelControl = new ChannelControl(this);
		add(channelControl, Image5DLayout.CHANNEL_SELECTOR);
        
		// Add slice selector	
        scrollbarsWL[3] = new ScrollbarWithLabel(Scrollbar.HORIZONTAL, 1, 1, 1, dimensions[3]+1, imp.getDimensionLabel(3));
		Scrollbars[3] = scrollbarsWL[3].getScrollbar();		
        if (i5d.getNSlices()>1) {
            add(scrollbarsWL[3], Image5DLayout.SLICE_SELECTOR);
        }

		// Add frame selector	
        scrollbarsWL[4] = new ScrollbarWithLabel(Scrollbar.HORIZONTAL, 1, 1, 1, dimensions[4]+1, imp.getDimensionLabel(4));
		Scrollbars[4] = scrollbarsWL[4].getScrollbar();		
        if (i5d.getNFrames()>1) {
            add(scrollbarsWL[4], Image5DLayout.FRAME_SELECTOR);
        }

		for(int i=3; i<nDimensions; ++i) {
			Scrollbars[i].addAdjustmentListener(this);
			int blockIncrement = dimensions[i]/10;
			if (blockIncrement<1) blockIncrement = 1;
			Scrollbars[i].setUnitIncrement(1);
			Scrollbars[i].setBlockIncrement(blockIncrement); 
		}			
		
		sliceSelector = Scrollbars[3];

		pack();

		
		isInitialized = true;	

		updateSliceSelector();	
		i5d.updateAndRepaintWindow();
		i5d.updateImageAndDraw();
        // Stop thread started in parent (StackWindow) constructor.
        done=true;
        thread.interrupt();       
        while(thread.isAlive()){}
        done=false;
        
		thread = new Thread(this, "SliceSelector");    
		thread.start();

        ImageJ ij = IJ.getInstance();
        
        // Handle key events for selecting c/z/t.
        // Image5DWindow has to receive the event before ImageJ, so
        // remove ImageJ as keylistener from this and all subcomponents,
        // add this as keylistener, and add ImageJ again.
        removeKeyListener(ij);
        ic.removeKeyListener(ij);

        
        addKeyListener(this);
        ic.addKeyListener(this);
        scrollbarsWL[3].addKeyListener(this);
        scrollbarsWL[4].addKeyListener(this);
        // Channel control takes care of itself.
        
        addKeyListener(ij);
        ic.addKeyListener(ij);
        scrollbarsWL[3].getScrollbar().addKeyListener(ij);
        scrollbarsWL[4].getScrollbar().addKeyListener(ij);

        
	}

//	public Image5D getImage5D() {
//		return (Image5D) imp;
//	}
	
	/** Handles changes in the scrollbars for z and t. 
	 */
	public synchronized void adjustmentValueChanged(AdjustmentEvent e) {
		if (!running2){
			
			for(int i=3; i<nDimensions; ++i) {
				if (e.getSource() == Scrollbars[i]) {
					positions[i] = Scrollbars[i].getValue();
                }
			}
			notify();
		}
	}

	public int getDisplayMode() {
		if (channelControl!=null) {
			return channelControl.getDisplayMode();
		}
		else {
			return -1;
		}
	}
	
	/** Handles change in ChannelControl. 
	 * Is called by ChannelControl without any events involved.
	 */
	public synchronized void channelChanged() {
		if (!running2){
			positions[2] = channelControl.getCurrentChannel();
		}
		
		notify();	
	}
	

	/** Updates the size and value of the stack and time scrollbar 
	 * and the size and value and other display properties of the channel control. */
	public void updateSliceSelector() {		
		// TODO: update other display properties of channel control
		if (isInitialized) {
			int[] newDimensions = imp.getDimensions();
			
			// update channel control
            if(newDimensions[2]>1 && dimensions[2]<=1 || newDimensions[2]<=1 && dimensions[2]>1) {
                channelControl.updateSelectorDisplay();
            }                
		    channelControl.setDisplayMode(i5d.getDisplayMode());
			channelControl.updateChannelSelector();
            
			// update z- and time control
			int max;
            
            // Display or remove slice and frame selectors
            if(newDimensions[3]>1 && dimensions[3]<=1) {
                add(scrollbarsWL[3], Image5DLayout.SLICE_SELECTOR);
            } else if (newDimensions[3]<=1 && dimensions[3]>1) {
                remove(scrollbarsWL[3]);
            }
            if(newDimensions[4]>1 && dimensions[4]<=1) {
                add(scrollbarsWL[4], Image5DLayout.FRAME_SELECTOR);
            } else if (newDimensions[4]<=1 && dimensions[4]>1) {
                remove(scrollbarsWL[4]);
            }         
            
            dimensions = newDimensions;
            
			for(int i=3; i<nDimensions; ++i) {
				max = Scrollbars[i].getMaximum();
				if (max!=(dimensions[i]+1)) {
					Scrollbars[i].setMaximum(dimensions[i]+1);
                    int blockIncrement = dimensions[i]/10;
                    if (blockIncrement<1) blockIncrement = 1;
                    Scrollbars[i].setBlockIncrement(blockIncrement); 
                }
				Scrollbars[i].setValue(((Image5D)imp).getCurrentPosition(i)+1);
			}
		}
	}

    public void drawInfo(Graphics g) {
   // TODO: add support for slice labels
        int TEXT_GAP = 0;
    	String s="";
		Insets insets = super.getInsets();
		Image5D img5 = (Image5D) imp;
		int[] dimensions = imp.getDimensions();
		Calibration cal = img5.getCalibration();
		
		// current position
		for (int i=2; i<img5.getNDimensions(); ++i){
			s += (img5.getDimensionLabel(i)).trim()+":";
			s += (img5.getCurrentPosition(i)+1);
			s += "/";
			s += dimensions[i];
			s += "; ";
		}    	
    	
    	// x/y size
    	if (cal.pixelWidth!=1.0 || cal.pixelHeight!=1.0)
    		s += IJ.d2s(imp.getWidth()*cal.pixelWidth,2) + "x" + IJ.d2s(imp.getHeight()*cal.pixelHeight,2)
 			+ " " + cal.getUnits() + " (" + imp.getWidth() + "x" + imp.getHeight() + "); ";
    	else
    		s += imp.getWidth() + "x" + imp.getHeight() + " pixels; ";
    	
    	// Data type and size in kBytes
		int size = 1;
		for (int i=0; i<img5.getNDimensions(); ++i) {
			size *= dimensions[i];
		}
		size /= 1024;
    	int type = imp.getType();
    	switch (type) {
	    	case ImagePlus.GRAY8:
	    		s += "8-bit";
	    		break;
	    	case ImagePlus.GRAY16:
	    		s += "16-bit grayscale";
				size *= 2;
	    		break;
	    	case ImagePlus.GRAY32:
	    		s += "32-bit grayscale";
				size *= 4;
	    		break;
    	}
    	if (size>=10000)    	
    		s += "; " + (int)Math.round(size/1024.0) + "MB";
    	else if (size>=1024) {
    		double size2 = size/1024.0;
    		s += "; " + IJ.d2s(size2,(int)size2==size2?0:1) + "MB";
    	} else
    		s += "; " + size + "K";
		g.drawString(s, 5, insets.top+TEXT_GAP);
    }
    
    public void paint(Graphics g) {
        //if (IJ.debugMode) IJ.log("wPaint: " + imp.getTitle());
        
        // Draw image info.
        g.setColor(Color.black);
        drawInfo(g);
        
        // Draw outlines of Canvasses and highlight current image.
        drawRectangles();
        //IJ.write(p + " " + d);
    }   
    
	public void run() {
		if (!isInitialized)
			return;
		while (!done) {
			synchronized(this) {
				try {wait(500);}
				catch(InterruptedException e) {}
			}
			if (done) return;

			for(int i=2; i<nDimensions; ++i) {
				if (positions[i]>0) {
					int p = positions[i];
					positions[i] = 0;
					if (p!=i5d.getCurrentPosition(i)+1) {
						i5d.setCurrentPosition(i, p-1);					
					}
				}
			}
			
		}
	}

    public ChannelControl getChannelControl() {
        return channelControl;
    }
    
//    // Handles key events to change c/z/t by keypresses
//    protected void processKeyEvent(KeyEvent e) {
//        if (e.getID() != KeyEvent.KEY_PRESSED || !(imp instanceof Image5D)) {
//            super.processKeyEvent(e);
//        } else {             
//            Image5D i5d = (Image5D)imp;
//            int code = e.getKeyCode();
//            boolean ctrlPressed = ((e.getModifiers()&KeyEvent.CTRL_MASK)!=0);
//            boolean shiftPressed = ((e.getModifiers()&KeyEvent.SHIFT_MASK)!=0);
//            
//            if (code==KeyEvent.VK_NUMPAD4 || (code==KeyEvent.VK_PAGE_DOWN&&shiftPressed )) {
//                // Numpad 4 / 6 and shift-pageDown / shift-pageUp for slice change
//                i5d.setSlice(i5d.getCurrentSlice()-1);                
//            } else if (code==KeyEvent.VK_NUMPAD6 || (code==KeyEvent.VK_PAGE_UP&&shiftPressed )) {
//                i5d.setSlice(i5d.getCurrentSlice()+1);    
//                
//            } else if (code==KeyEvent.VK_NUMPAD7 || (code==KeyEvent.VK_PAGE_DOWN&&ctrlPressed )) {
//                // Numpad 7 / 9 and ctrl-pageDown / ctrl-pageUp for frame change
//                i5d.setFrame(i5d.getCurrentFrame()-1);                
//            } else if (code==KeyEvent.VK_NUMPAD9 || (code==KeyEvent.VK_PAGE_UP&&ctrlPressed )) {
//                i5d.setFrame(i5d.getCurrentFrame()+1);  
//                
//            } else if (code==KeyEvent.VK_NUMPAD1 || (code==KeyEvent.VK_PAGE_DOWN )) {
//                // Numpad 4 / 6 and pageDown / pageUp (w/o modifier) for channel change
//                i5d.setChannel(i5d.getCurrentChannel()-1);                
//            } else if (code==KeyEvent.VK_NUMPAD3 || (code==KeyEvent.VK_PAGE_UP )) {
//                i5d.setChannel(i5d.getCurrentChannel()+1);   
//                
//            } else {
//                // If none of these combinations is pressed, pass on the event.
//                super.processKeyEvent(e);
//            }
//        }
//    }
    
    // methods of the Keylistener interface

    public void keyPressed(KeyEvent e) {        
        if (e.getID() == KeyEvent.KEY_PRESSED && (imp instanceof Image5D)) {             
            Image5D i5d = (Image5D)imp;
            int code = e.getKeyCode();
            boolean ctrlPressed = ((e.getModifiers()&KeyEvent.CTRL_MASK)!=0);
            boolean shiftPressed = ((e.getModifiers()&KeyEvent.SHIFT_MASK)!=0);
                        
            // Process defined key events. Make them unusable for ImageJ by setting the 
            // KeyCode to CHAR_UNDEFINED
            if (code==KeyEvent.VK_NUMPAD4 || (code==KeyEvent.VK_PAGE_DOWN&&shiftPressed ) 
                    || code==KeyEvent.VK_LESS || code==KeyEvent.VK_COMMA) {
                // Numpad 4 / 5 and shift-pageDown / shift-pageUp for slice change
                i5d.setSlice(i5d.getCurrentSlice()-1);      
                e.setKeyCode(KeyEvent.CHAR_UNDEFINED);
            } else if (code==KeyEvent.VK_NUMPAD5 || (code==KeyEvent.VK_PAGE_UP&&shiftPressed )
                    || code==KeyEvent.VK_GREATER || code==KeyEvent.VK_PERIOD) {
                i5d.setSlice(i5d.getCurrentSlice()+1);   
                e.setKeyCode(KeyEvent.CHAR_UNDEFINED); 
                
            } else if (code==KeyEvent.VK_NUMPAD1 || (code==KeyEvent.VK_PAGE_DOWN&&ctrlPressed )) {
                // Numpad 1 / 2 and ctrl-pageDown / ctrl-pageUp for frame change
                i5d.setFrame(i5d.getCurrentFrame()-1); 
                e.setKeyCode(KeyEvent.CHAR_UNDEFINED);               
            } else if (code==KeyEvent.VK_NUMPAD2 || (code==KeyEvent.VK_PAGE_UP&&ctrlPressed )) {
                i5d.setFrame(i5d.getCurrentFrame()+1);  
                e.setKeyCode(KeyEvent.CHAR_UNDEFINED);
                
            } else if (code==KeyEvent.VK_NUMPAD7 || (code==KeyEvent.VK_PAGE_DOWN )) {
                // Numpad 7 / 8 and pageDown / pageUp (w/o modifier) for channel change
                i5d.setChannel(i5d.getCurrentChannel()-1);  
                e.setKeyCode(KeyEvent.CHAR_UNDEFINED);              
            } else if (code==KeyEvent.VK_NUMPAD8 || (code==KeyEvent.VK_PAGE_UP )) {
                i5d.setChannel(i5d.getCurrentChannel()+1);   
                e.setKeyCode(KeyEvent.CHAR_UNDEFINED);               
            }
        }
    }

    public void keyReleased(KeyEvent e) {
    }
    public void keyTyped(KeyEvent e) {
    }
    
	
	protected void drawRectangles() {
		Graphics g = this.getGraphics();
        LayoutManager layout = getLayout();
        if (layout != null && layout instanceof Image5DLayout) {      
            
            // Paint over all old rubbish.
            g.setColor(Color.white);
            Rectangle coBounds = ((Image5DLayout)layout).getContentBounds(); 
            if (coBounds != null) {
                g.fillRect(coBounds.x, coBounds.y, coBounds.width, coBounds.height);
            }
            
            // Draw rectangles.
            // If in gallery mode, rectangle is red for current channel and
            // black for the others.
            Rectangle caBounds = ((Image5DLayout)layout).getCanvasBounds(0);
            if (caBounds != null) {
                g.setColor(Color.black);
                g.drawRect(caBounds.x-1, caBounds.y-1, caBounds.width+1, caBounds.height+1);
            }
            // Red rectangle for current channel.
//            if (caBounds != null) {
//                g.setColor(Color.red);
//                g.drawRect(caBounds.x-1, caBounds.y-1, caBounds.width+1, caBounds.height+1);
//                g.drawRect(caBounds.x-2, caBounds.y-2, caBounds.width+3, caBounds.height+3);
//                g.drawRect(caBounds.x-3, caBounds.y-3, caBounds.width+5, caBounds.height+5);
//            }
        }
	}

}